"""Code dealing directly with the text user interface.

Currently this ties in heavily with curses. Eventually, curses will be
abstracted out.
"""

import curses
import os
import string
import sys
from .commands import SINGLE_KEY_COMMANDS, COMMANDS
from . import env, nums, curses_stuff

def format_stack(stack, recurse=True, firstcall=True):
    """Converts a single stack to a list of lines of text.
    If the stack has a parent and recurse is true, draw the parents
    as well.

    Also returns a list of x coordinates where dividers should be
    drawn.
    """

    lines = []

    for i, item in enumerate(stack):
        formatted = env.FORMAT.format(item)
        stack_n = i

        margin = ("%3d  " % stack_n) if firstcall else "| "

        if hasattr(item, "tag"):
            text = "%20s : %s" % (formatted, item.tag)
        else:
            text = "%20s" % formatted

        lines.append(margin + text)

    lines = lines[::-1]

    if recurse and stack.parent is not None:
        parent_lines, parent_dividers = format_stack(stack.parent, firstcall=False)

        # Line up lines at the bottom
        lines.extend([""] * (len(parent_lines) - len(lines)))
        parent_lines.extend([""] * (len(lines) - len(parent_lines)))

        # Concatenate
        concat_len = 35 if firstcall else 15

        joined_lines = []
        for i, j in zip(lines, parent_lines):
            if len(i) < concat_len:
                i += " " * (concat_len - len(i))
            i = i[:concat_len]
            joined_lines.append(i + j)
        return joined_lines, [concat_len] + [i + concat_len for i in parent_dividers]

    return lines, []


def draw_stack(stack):
    """Draw the stack.
    """

    lines, dividers = format_stack(stack)

    curses_stuff.draw_text(env.STDSCR, lines, offset=-1, xoffset=0, padding=2)

    for xoff in dividers:
        curses_stuff.draw_vline(env.STDSCR, xoff, 2, -1)

    curses.doupdate()

def get_single_key_command(key):
    """
    Return a single-key command for the key, or None.
    """
    if key > 127:
        lookup_key = key
    else:
        try:
            lookup_key = chr(key)
        except ValueError:
            lookup_key = key
    return SINGLE_KEY_COMMANDS.get(lookup_key, None)

def do_command(stack, cmd, arg):
    """ Execute the command on the stack in an exception wrapper.

    Returns None if no exception, or sys.exc_info() if there was one.
    """

    # pylint: disable=broad-except
    try:
        stack.undopush()
        rtn = cmd(stack, arg)
    except Exception:
        stack.undo()
        return sys.exc_info()
    else:
        assert rtn is None
        return None

ENTER_KEYS = [curses.KEY_ENTER, 10, 13]
BACKSPACE_KEYS = [curses.KEY_BACKSPACE, 127, 8]

def do_edit(key, strbuf):
    """
    Line-edit.
    Returns edit_type, new_strbuf
        edit_type: one of "BACKSPACE", "ENTER", None
    """

    if key in BACKSPACE_KEYS:
        return "BACKSPACE", strbuf[:-1]

    if key in ENTER_KEYS:
        return "ENTER", strbuf

    try:
        c = chr(key)
    except ValueError:
        c = ""

    return None, strbuf + c

def do_exit(stack, extra_text=None):
    """Perform the exit action.

    In this case, the stack is formatted to stdout,
    then we exit.
    """

    def exit_helper():
        """Format extra text and the stack to stdout."""
        if "LERPN_TEST_PRINT" in os.environ:
            print("BEGIN_PRINT")
            if extra_text is not None:
                print(extra_text)
            for item in stack:
                natural = nums.NaturalMode.format(item)
                print("%s" % (natural))

    # Using atexit allows us to register the function to be called after
    # curses.wrapper() performs cleanup
    import atexit
    atexit.register(exit_helper)

    sys.exit(0)

class Prompt(object):
    """State machine for user input prompt"""

    def __init__(self):
        self.state = self.state_firstchar
        self.strbuf = ""
        self.actions = []
        self.key = None
        self.char = None
        self.edit_type = None
        self.exception = None

    def loop(self, stack):
        """Displays the prompt and executes a user action from it.
        stack:  buffer current UndoStack
        """
        if self.exception is not None:
            env.LAST_ERROR = self.exception
            env.STDSCR.addstr(0, 0, "ERROR: %s" % str(self.exception[1]))
            self.exception = None

        self.state = self.state_firstchar
        self.strbuf = ""

        while self.state is not None:
            self.cycle(stack)

    def cycle(self, stack):
        """Perform one cycle of the state machine. Not generally called externally."""
        self.actions = []
        self.exception = None

        if self.do_io(stack):
            return

        if self.state is not None:
            self.state()

        self.do_actions(stack)

    def do_io(self, stack):
        """Handle input/output. Returns True if this cycle should be skipped."""

        draw_stack(stack)
        prompt_line = "? " + self.strbuf
        curses_stuff.draw_textline(env.STDSCR, prompt_line, -1, 0, True)
        curses.doupdate()

        self.key, self.char = curses_stuff.getkey(env.STDSCR)

        if self.key == curses.KEY_F5:
            self.actions.append(('push', str(self.state)))
        elif self.key == curses.KEY_RESIZE:
            env.STDSCR.clear()
            return True

        self.edit_type, self.strbuf = do_edit(self.key, self.strbuf)
        return False

    def do_actions(self, stack):
        """Execute any actions accumulated over a machine cycle."""
        for cmd, arg in self.actions:
            if cmd == "parse":
                # pylint: disable=broad-except
                try:
                    parsed = nums.num_parser(arg)
                except Exception:
                    self.exception = sys.exc_info()
                else:
                    stack.undopush()
                    stack.append(parsed)
            elif cmd == "push":
                stack.append(arg)
            elif cmd == "undopush":
                stack.undopush()
            else:
                self.exception = do_command(stack, cmd, arg)


    def state_firstchar(self):
        """State 'firstchar': look at the first character typed and determine the kind of input."""
        # Try single-key commands first
        cmd = get_single_key_command(self.key)
        if cmd is not None:
            self.actions.append((cmd, None))
            self.state = None

        elif self.char is None:
            self.state = self.state_firstchar

        elif self.char in string.digits + "_.":
            self.state = self.state_numeric

        elif self.char == "'":
            self.state = self.state_command

        elif self.char == '"':
            self.state = self.state_string

        else:
            self.strbuf = ""

    def state_numeric(self):
        """State 'numeric': number entry"""

        # When typing a number, it is sometimes necessary to forbid
        # the interpretation of single-key commands:

        cmdblock_tag = ('"' in self.strbuf)
        cmdblock_sci = (self.strbuf[-2:-1] in "ej")
        cmdblock_hex = (self.strbuf[0:2] == "0h" and
                        self.char is not None and
                        self.char in "abcdef")

        cmdblock = cmdblock_tag or cmdblock_sci or cmdblock_hex

        if cmdblock or self.edit_type in ("BACKSPACE", "ENTER"):
            cmd = None
        else:
            cmd = get_single_key_command(self.key)

        if not self.strbuf:
            self.state = None

        elif self.char == "'":
            self.state = self.state_command
            self.actions.append(("parse", self.strbuf[:-1]))
            self.strbuf = "'"

        elif self.edit_type == "ENTER":
            self.state = None
            self.actions.append(("parse", self.strbuf))
            self.strbuf = ""

        elif cmd is not None:
            self.state = None
            self.actions.append(("parse", self.strbuf[:-1]))
            self.actions.append((cmd, None))
            self.strbuf = ""

    def state_command(self):
        """State 'command': named command entry"""
        if self.edit_type == "ENTER":
            cmdname, _, arg = self.strbuf.partition(" ")
            cmd = COMMANDS.get(cmdname, None)
            if cmd is not None:
                self.actions.append((cmd, arg))
            elif len(self.strbuf):
                try:
                    raise KeyError("Command not found: %s" % self.strbuf)
                except KeyError:
                    self.exception = sys.exc_info()

            self.state = None

        if not self.strbuf:
            self.state = None

    def state_string(self):
        """State 'string': string entry"""
        if not self.strbuf:
            self.state = None
        elif self.edit_type == "ENTER":
            self.actions.append(("push", self.strbuf.lstrip('"')))
            self.state = None




"""lerpn application module

To run lerpn, call main().
"""

__version__ = "2.0"

# Written in 2015 by Christopher Pavlina.
###############################################################################
#
# This software is free. Do whatever the fuck you want with it. Copy it, change
# it, steal it, strip away this notice, print it and tear it into confetti to
# throw at random dudes on the street. Really, I don't care.

# pylint: disable=wrong-import-position
import curses
import os
import signal
import sys
from . import nums, env, tui

def wrapped_main(stdscr):
    """main function to be wrapped by curses.wrapper
    """
    curses.nonl()
    env.STDSCR = stdscr
    env.STACK = nums.UndoStack()
    prompt = tui.Prompt()

    # Register exit handler
    # pylint: disable=unused-argument
    def exit_handler(signum, frame):
        """Call the lerpn cleanup/exit function, to be used as a signal handler"""
        tui.do_exit(env.STACK)

    signal.signal(signal.SIGINT, exit_handler)

    while True:
        stdscr.clear()
        prompt.loop(env.STACK)

def main():
    """call this to launch lerpn
    """
    debug = os.getenv("DEBUG_LERPN")
    if debug is not None and int(debug):
        import rpdb
        rpdb.set_trace()

    if "--help" in sys.argv or "-h" in sys.argv:
        print("usage: %s [--version] [--help]" % os.path.basename(sys.argv[0]))
        print("")
        print("lerpn (Linux Engineering RPN calculator)")
        print("Run with no arguments to start the calculator. Once started, try")
        print("pressing ? to see a list of commands.")
        return

    if "--version" in sys.argv:
        print("lerpn version " + __version__)
        return

    curses.wrapper(wrapped_main)

def gen_commands_md():
    """generate a Markdown reference of all commands"""

    from . import commands, curses_stuff

    print("# Quick Command Reference")
    print()
    print("This list is auto-generated from the code, and contains the same")
    print("text you would see in the in-program help box.")
    print()
    print("## Single-key commands")
    print()
    print("Key | Description")
    print("--- | -----------")
    for cmd in commands.SINGLE_KEY_COMMANDS:
        cmd_name = curses_stuff.KEY_TO_NAME.get(cmd, "`%s`" % cmd)
        cmd_function = commands.SINGLE_KEY_COMMANDS[cmd]
        if cmd_function is None:
            continue
        cmd_desc = cmd_function.__doc__.partition("\n")[0]
        print("%s | %s" % (cmd_name, cmd_desc))
    print()

    print("## Long commands")
    print()
    print("Name | Description")
    print("---- | -----------")
    for cmd in commands.COMMANDS:
        if isinstance(cmd, int):
            cmd_name = ""
            cmd_desc = "**%s**" % commands.COMMANDS[cmd]
        else:
            cmd_name = "`%s`" % cmd
            cmd_desc = commands.COMMANDS[cmd].__doc__.partition("\n")[0]
        print("%s | %s" % (cmd_name, cmd_desc))
    print()

if __name__ == "__main__":
    main()



"""Curses utilities.
"""

import curses

# Reference of nonprintables to printable names
KEY_TO_NAME = {
    "\x7f": "[0x7f]",
    "\x08": "[0x08]",
    curses.KEY_BACKSPACE: "[bksp]",
    "\r": "[CR]",
    "\n": "[LF]",
    curses.KEY_ENTER: "[enter]",
    curses.KEY_UP: "[up]",
    curses.KEY_DOWN: "[down]",
    curses.KEY_LEFT: "[left]",
    curses.KEY_RIGHT: "[right]",
    curses.KEY_PPAGE: "[pgup]",
    curses.KEY_NPAGE: "[pgdn]",
    curses.KEY_HOME: "[home]",
    curses.KEY_END: "[end]",
}

# pylint: disable=too-many-instance-attributes
class CursesScrollBox(object):
    """Displays some text in a scroll box."""

    def __init__(self, parent, width, height):
        """Initialize.

        @param width - desired width: positive for absolute width, negative
                    for margin around window edge
        @param height - desired height: positive for absolute height,
                    negative for margin around window edge
        """
        self.parent = parent
        self.width = width
        self.height = height
        self.title = ""
        self.text = ""
        self.vpos = 0
        self.hpos = 0

        self._width = None
        self._height = None
        self._xoff = None
        self._yoff = None

    def set_text(self, text):
        """Set the text displayed in the scroll box. This can be a string, which
        will be split into lines, or a list of lines"""

        # Split the text?
        if isinstance(text, str):
            text_split = text.split("\n")
        elif isinstance(text, list):
            text_split = text
        else:
            raise TypeError("Text must be list or str")

        # Tab-expand the text so it horizontally scrolls nicely
        lines = []
        for line in text_split:
            new_line = []
            col = 0
            for char in line:
                col += 1
                if char == "\t":
                    n_spaces = 8 - (col - 1) % 8
                    new_line.extend([" "] * n_spaces)
                    col += n_spaces - 1
                else:
                    new_line.append(char)
            lines.append(''.join(new_line))
        self.text = lines

    def set_title(self, title):
        """Set the title to be displayed at the top."""
        self.title = title

    def show(self):
        """
        Display the window on the curses display
        """

        curses.curs_set(0)
        curses.doupdate()
        self._set_dimensions(None)
        window = curses.newwin(self._height, self._width, self._yoff, self._xoff)
        window.keypad(1)

        cycle = True
        while cycle:
            self._draw(window)
            cycle = self._nav(window)

        curses.curs_set(1)


    def _set_dimensions(self, window):
        """
        Configure _width, _height, _xoff, _yoff from screen dimensions.
        """
        lines, cols = self.parent.getmaxyx()

        if self.width < 0:
            self._width = cols + self.width
        else:
            self._width = self.width
        if self.height < 0:
            self._height = lines + self.height
        else:
            self._height = self.height
        self._xoff = cols // 2 - self._width // 2
        self._yoff = lines // 2 - self._height // 2

        if window is not None:
            window.resize(self._height, self._width)
            try:
                window.move(self._yoff, self._xoff)
            except curses.error:
                pass

    def _clamp_offsets(self):
        """
        Clamp vpos and hpos into the display range
        """

        longest_line = max(len(i) for i in self.text)
        yscroll_limit = max(0, len(self.text) - self._height // 2)
        xscroll_limit = max(0, longest_line - self._width // 2)

        self.vpos = max(0, self.vpos)
        self.vpos = min(self.vpos, yscroll_limit)

        self.hpos = max(0, self.hpos)
        self.hpos = min(self.hpos, xscroll_limit)


    def _draw(self, window):
        """
        Draw the scroll box.
        """

        window.clear()
        try:
            window.addstr(1, 2, self.title)
        except curses.error:
            pass

        self._clamp_offsets()

        view_lines = self.text[self.vpos:self.vpos+self._height - 4]

        try:
            for idx, line in enumerate(view_lines):
                window.addstr(idx + 3, 2, line[self.hpos:self._width - 2 + self.hpos])
        except curses.error:
            pass

        try:
            if self.vpos > 0:
                window.addstr(2, self._width - 3, "+")
            if self.vpos + self._height < len(self.text) + 4:
                window.addstr(self._height - 2, self._width - 3, "+")
        except curses.error:
            pass

        window.border()

    def _nav(self, window):
        """Handle the navigation keys. Return True to continue for another cycle,
        or False to exit."""

        try:
            key = window.getch()
        except KeyboardInterrupt:
            key = None

        if key in (curses.KEY_UP, ord('k')):
            self.vpos -= 1
        elif key in (curses.KEY_DOWN, ord('j')):
            self.vpos += 1
        elif key in (curses.KEY_LEFT, ord('h')):
            self.hpos -= 5
        elif key in (curses.KEY_RIGHT, ord('l')):
            self.hpos += 5

        elif key == curses.KEY_PPAGE:
            self.vpos -= (self._height // 2)
        elif key == curses.KEY_NPAGE:
            self.vpos += (self._height // 2)

        elif key == curses.KEY_HOME:
            self.vpos = 0
        elif key == curses.KEY_END:
            self.vpos = len(self.text)

        elif key == curses.KEY_RESIZE:
            self._set_dimensions(window)
            self.parent.clear()
            self.parent.refresh()

        else:
            return False

        return True

def draw_text(window, lines, offset, xoffset, padding):
    """Draw a list of text lines onto the window.

    @param window - curses window
    @param lines - list of lines
    @param offset - line number to start at. Negative to start at bottom
    @param xoffset - horizontal offset
    @param padding - distance from other end at which to stop
    """

    w_height, w_width = window.getmaxyx()
    lines_wclamp = [i[:w_width] for i in lines]
    lines_hwclamp = lines_wclamp[:(w_height - abs(offset) - padding)]

    for idx, line in enumerate(lines_hwclamp):
        if offset < 0:
            yoffset = w_height - idx + offset - 1
        else:
            yoffset = idx + offset
        window.addstr(yoffset, xoffset, line)

def draw_vline(window, xpos, ystart, height):
    """Draw a vertical line onto the window

    @param window - curses window
    @param xpos - x coordinate (pos from left, neg from right)
    @param ystart - y coordinate (pos from top, neg from bottom)
    @param height - line height (pos absolute, neg relative to window height)
    """

    w_height, w_width = window.getmaxyx()

    xpos %= w_width
    height %= w_height

    if ystart < 0:
        ystart = w_height - height + ystart

    # pylint: disable=no-member
    # ACS_VLINE is populated at load
    window.vline(ystart, xpos, curses.ACS_VLINE, height)

def draw_textline(window, line, yoff, xoff, clearline=False):
    """Draw a line of text onto the window

    @param window - curses window
    @param line - line of text
    @param yoff - y offset, pos from top, neg from bottom
    @param xoff - x offset, pos from left, neg from right
    @param clearline - True to clear the rest of the line
    @return True on success (can fail if yoff/xoff is invalid)
    """

    w_height, w_width = window.getmaxyx()

    yoff %= w_height
    xoff %= w_width

    try:
        window.move(yoff, xoff)
    except curses.error:
        return False

    window.addstr(line)
    if clearline:
        window.clrtoeol()

    return True

def getkey(window):
    """Get a keypress from the window.

    @return keycode, character  - character will be None if there is no match
    """

    key = window.getch()
    try:
        char = chr(key)
    except ValueError:
        char = None
    return key, char




"""Extra mathematical functions."""

import cmath
import math

def degree_trig(func):
    """Create a unary trig operator that accepts degrees."""
    def wrapper(value):
        """Wrapper function: trig(degrees)"""
        value_deg = value * math.pi / 180
        return func(value_deg)
    return wrapper

def degree_invtrig(func):
    """Create a unary inverse trig operator that returns degrees.
    """
    def wrapper(value):
        """Wrapper function: rad2deg(trig(value))"""
        return func(value) * 180 / math.pi
    return wrapper

def autoc(fname):
    """Return a wrapper function that calls Call math.<fname> for floats and
    ints, cmath.<fname> for complex"""

    def wrapper(n, *args):
        """Wrapper function: math.fname(float), cmath.fname(complex)"""
        if isinstance(n, complex):
            return getattr(cmath, fname)(n, *args)
        else:
            return getattr(math, fname)(n, *args)
    return wrapper

def clamp(lower, n, upper):
    """Return n clamped to the range lower <= n <= upper"""

    return max(lower, min(n, upper))

# From https://code.activestate.com/recipes/66472/
# Originally by Dinu Gherman (first version) and Walter Brunswick (improved)
# Python Software Foundation license
def frange(end, start=0, inc=0):
    """A range function that accepts float increments."""

    if not start:
        start = end + 0.0
        end = 0.0
    else: end += 0.0

    if not inc:
        inc = 1.0
    count = int(math.ceil((start - end) / inc))

    L = [None] * count

    L[0] = end
    for i in range(1, count):
        L[i] = L[i-1] + inc
    return L



"""All the calculator commands.

This file contains the commands made available to the user. Single-key
commands are in the SINGLE_KEY_COMMANDS OrderedDict, and full-name
commands are in the COMMANDS OrderedDict.
"""

# pylint: disable=missing-docstring,unused-argument

import cmath
import collections
import curses
import math
import os
import subprocess
import sys
import traceback

from . import nums, env, curses_stuff
from .extramath import *

###############################################################################
# COMMAND GENERATORS
#
# Use these to generate common cases, rather than implementing from scratch.

def BINARY(func, docstring=None):
    """Create a binary operator from a lambda function."""
    def cmd(stack, arg):
        y_val = stack.pop()
        x_val = stack.pop()
        stack.append(func(x_val, y_val))

    cmd.__doc__ = docstring or func.__doc__
    assert cmd.__doc__

    return cmd

def UNARY(func, docstring=None, autostack=True):
    """Create a unary operator from a lambda function.
    The unary operators know how to handle tagged numbers - they reapply the
    same tag after performing the operation.

    expect_stack: if True (default), automatically expand func(stack) to
    [func(i) for i in stack]. if False, pass the stack to func like any
    other number."""
    def cmd_one_value(value):
        if isinstance(value, nums.Tagged):
            newtagged = nums.Tagged(func(value.num), value.tag)
            return newtagged
        else:
            return func(value)

    def cmd_noauto(stack, arg):
        stack.append(cmd_one_value(stack.pop()))

    def cmd_auto(stack, arg):
        value = stack.pop()
        if isinstance(value, nums.UndoStack):
            newstack = nums.UndoStack([cmd_one_value(i) for i in value])
            newstack.parent = stack
            stack.append(newstack)
        else:
            stack.append(cmd_one_value(value))

    if autostack:
        cmd = cmd_auto
    else:
        cmd = cmd_noauto

    cmd.__doc__ = docstring or func.__doc__
    assert cmd.__doc__

    return cmd

def CONST(val, descr):
    """Return a command function to push a constant"""

    def cmd(stack, arg):
        stack.append(val)
    cmd.__doc__ = descr
    return cmd

###############################################################################
# COMMAND HELPERS
def untagged(num):
    """return a number without its tag, if it is tagged"""
    if isinstance(num, nums.Tagged):
        return num.num
    else:
        return num

###############################################################################
# MATH COMMANDS

def _reim(stack, arg):
    """split into real and imaginary parts"""
    num = untagged(stack.pop())
    stack.append(num.real)
    stack.append(num.imag)

def _polar(stack, arg):
    """split into absolute value and angle"""
    num = untagged(stack.pop())
    stack.append(abs(num))
    stack.append(cmath.phase(num))

def _fcart(stack, arg):
    """make complex from Cartesian real, imag"""
    imag = untagged(stack.pop())
    real = untagged(stack.pop())

    stack.append(complex(real, imag))

def _fpolar(stack, arg):
    """make complex from polar abs val, angle(rad)"""
    angle = untagged(stack.pop())
    absval = untagged(stack.pop())

    stack.append(cmath.rect(absval, angle))

def _pvar(stack, arg):
    """population variance"""
    nest = stack.pop()
    mean = sum(nest) / float(len(nest))
    sdiff = [(i - mean) ** 2 for i in nest]
    pvar = sum(sdiff) / float(len(sdiff))
    stack.append(pvar)

def _pdev(stack, arg):
    """population std deviation"""
    _pvar(stack, arg)
    stack.append(stack.pop() ** 0.5)

def _svar(stack, arg):
    """sample variance"""
    nest = stack.pop()
    mean = sum(nest) / float(len(nest))
    sdiff = [(i - mean) ** 2 for i in nest]
    svar = sum(sdiff) / float(len(sdiff) - 1)
    stack.append(svar)

def _sdev(stack, arg):
    """sample std deviation"""
    _svar(stack, arg)
    stack.append(stack.pop() ** 0.5)

###############################################################################
# TAG COMMANDS

def _tag(stack, arg):
    """attach tag to number"""

    if arg:
        tag = arg.strip()
    else:
        tag = stack.pop()

    num = stack.pop()
    if not isinstance(tag, str):
        raise TypeError("to create tagged number, give Y: number, X: string")

    elif not isinstance(num, float) and not isinstance(num, nums.UndoStack):
        raise TypeError("to create tagged number, give Y: number, X: string ")

    stack.append(nums.Tagged(num, tag))

def _untag(stack, arg):
    """unTag"""
    stack.append(untagged(stack.pop()))

################################################################################
# STACK/SHELL COMMANDS

def _drop(stack, arg):
    """drop"""
    stack.pop()

def _dup(stack, arg, n=-1):
    """dup"""

    # @param n - which index to dup. This allows _dup to be wrapped by other
    # commands like _get and avoid some duplication.

    if n == -1:
        # pop instead of access so we get the same, consistent error message if
        # there isn't one.
        item = stack.pop(n)
        stack.append(item)
    else:
        item = stack[n]

    if isinstance(item, nums.UndoStack):
        new_stack = nums.UndoStack(item)
        new_stack.parent = stack
        stack.append(new_stack)
    elif isinstance(item, nums.Tagged) and isinstance(item.num, nums.UndoStack):
        new_stack = nums.UndoStack(item.num)
        new_stack.parent = stack
        stack.append(nums.Tagged(new_stack, item.tag))
    else:
        stack.append(item)

def _rotup(stack, arg):
    """rotate up"""
    stack.append(stack.pop(0))

def _rotdown(stack, arg):
    """rotate down"""
    stack.insert(0, stack.pop(-1))

def _pull(stack, arg):
    """move item from index to front"""
    idx = int(stack.pop())
    stack.append(stack.pop(idx))

def _push(stack, arg):
    """move item from front to index"""
    idx = int(stack.pop())
    stack.insert(idx, stack.pop())

def _undo(stack, arg):
    """undo"""
    # Drop the undopush that was done for /this/ command
    stack.replace_redo()
    stack.drop()
    stack.undo()

def _redo(stack, arg):
    """redo"""
    # Drop the undopush that was done for /this/ command
    stack.replace_redo()
    stack.drop()
    stack.redo()

def _xchg(stack, arg):
    """exchange/swap"""
    x = stack.pop()
    y = stack.pop()
    stack.append(x)
    stack.append(y)

def _get(stack, arg):
    """dup item by number"""
    stack_idx = clamp(0, int(untagged(stack.pop())), len(stack) - 1)
    _dup(stack, arg, n=stack_idx)

def _sh(stack, arg):
    """run shell command, $N = number"""
    assert arg, "argument expected!"
    newenv = dict(os.environ)
    if not stack:
        newenv["N"] = "0"
    else:
        num = untagged(stack[-1])
        if isinstance(num, float):
            newenv["N"] = "%g" % num
        elif isinstance(num, complex):
            newenv["N"] = "%g+%gj" %(num.real, num.imag)

    try:
        proc = subprocess.Popen(
            arg, shell=True, env=newenv, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        data = proc.stdout.read()
        proc.wait()
    except KeyboardInterrupt:
        data = b""

    if sys.version_info >= (3, 0):
        data = data.decode("utf8")

    match = nums.FLOAT_RE.search(data)
    if match is not None:
        try:
            stack.append(float(match.group(0)))
        except ValueError:
            pass

    scrollbox = curses_stuff.CursesScrollBox(env.STDSCR, -4, -4)
    scrollbox.set_title("COMMAND OUTPUT")
    scrollbox.set_text(data)
    scrollbox.show()

def _newstack(stack, arg):
    """push a new nested stack"""
    new_stack = nums.UndoStack()
    new_stack.parent = stack
    stack.append(new_stack)
    env.STACK = new_stack

def _enterstack(stack, arg):
    """enter a nested stack"""
    child = untagged(stack[-1])
    if not isinstance(child, nums.UndoStack):
        raise TypeError("can only enter a nested stack, not a value")
    env.STACK = child

    # Remove extra undo item
    stack.undo()

def _leavestack(stack, arg):
    """return to the parent stack"""
    if stack.parent is None:
        raise Exception("cannot move above the root stack")
    env.STACK = stack.parent

    # Remove extra undo item
    stack.undo()

def _merge(stack, arg):
    """merge a nested stack into this stack"""
    nested = stack.pop()
    if not isinstance(nested, nums.UndoStack):
        raise TypeError("can only merge stack, not value")
    for i in nested:
        if isinstance(i, nums.UndoStack):
            new_i = nums.UndoStack(i)
            new_i.parent = stack
            stack.append(new_i)
        else:
            stack.append(i)

def _flatten(stack, arg):
    """flatten the current stack"""

    def flatten_recursive(items, target):
        for i in items:
            if isinstance(i, nums.UndoStack):
                flatten_recursive(i, target)
            else:
                target.append(i)

    items = stack[:]
    del stack[:]
    flatten_recursive(items, stack)

def _seq(stack, arg):
    """seq start stop step"""

    args = [float(i) for i in arg.split()]
    if len(args) != 3:
        raise Exception("expected args: start stop step")

    start, stop, step = args

    newstack = nums.UndoStack(frange(start, stop, step))
    newstack.parent = stack
    stack.append(newstack)


################################################################################
# DISPLAY/FORMATTING/SETTINGS/HELP COMMANDS

def _help(stack, arg):
    """display help"""
    scrollbox = curses_stuff.CursesScrollBox(env.STDSCR, -4, -4)
    scrollbox.set_title("HELP")

    lines_left = []
    lines_right = []

    for key in COMMANDS.keys():
        val = COMMANDS[key]
        if isinstance(key, int):
            lines_left.append(" " * 26)
            lines_left.append("%-26s" % val)
        else:
            doc = val.__doc__
            if doc is None:
                doc = "NODOC: " + key
            lines_left.append("%-10s %-15s" % (key, doc.partition("\n")[0]))

    for key in SINGLE_KEY_COMMANDS.keys():
        key_name = curses_stuff.KEY_TO_NAME.get(key, key)
        val = SINGLE_KEY_COMMANDS[key]
        if val is None:
            continue
        lines_right.append("%-10s %-15s" % (key_name, val.__doc__.partition("\n")[0]))

    if len(lines_left) < len(lines_right):
        lines_left.extend([""] * (len(lines_right) - len(lines_left)))
    if len(lines_left) > len(lines_right):
        lines_right.extend([""] * (len(lines_left) - len(lines_right)))

    lines = [i + "    " + j for i, j in zip(lines_left, lines_right)]
    scrollbox.set_text(lines)

    scrollbox.show()

def _exc(stack, arg):
    """display the latest exception"""
    if env.LAST_ERROR is None:
        return
    last_bt = traceback.format_exception(*env.LAST_ERROR)
    scrollbox = curses_stuff.CursesScrollBox(env.STDSCR, -4, -4)
    scrollbox.set_title("LAST BACKTRACE")
    scrollbox.set_text(last_bt)
    scrollbox.show()

def _repr(stack, arg):
    """debug: repr(x)"""
    stack.append(repr(stack.pop()))

def _str(stack, arg):
    """debug: stringify"""
    stack.append(env.FORMAT.format(stack.pop()))

def _nat(stack, arg):
    """natural mode"""
    env.FORMAT = nums.NaturalMode

def _fix(stack, arg):
    """fixed mode, digits = 6 or command argument"""
    env.FORMAT = nums.FixMode

    if arg is not None and arg.strip():
        nums.FixMode.digits = int(arg.strip())
    else:
        nums.FixMode.digits = 6

def _eng(stack, arg):
    """engineering mode, sig figs = 7 or command argument"""
    env.FORMAT = nums.EngMode

    if arg is not None and arg.strip():
        nums.EngMode.sigfigs = int(arg.strip())
    else:
        nums.EngMode.sigfigs = 7

def _sci(stack, arg):
    """scientific mode"""
    env.FORMAT = nums.SciMode

def _hex(stack, arg):
    """hexadecimal display mode"""
    env.FORMAT = nums.HexMode

def _oct(stack, arg):
    """octal display mode"""
    env.FORMAT = nums.OctMode

def _bin(stack, arg):
    """binary display mode"""
    env.FORMAT = nums.BinMode


################################################################################
# ADDING SINGLE-LETTER COMMANDS
#
# Insert a tuple into the OrderedDict below:
# (key, command)
#
# key:      the key returned by curses. For printables, just the character.
# command:  a function that takes the current UndoStack, operates on it,
#               and returns nothing. Any exceptions thrown in here will be
#               caught and handled, so don't worry about those.
#
# The command must have a docstring. The first line of this will be used as its
# help text.
#
# Standard unary and binary operators can be made with UNARY() and BINARY(),
# which accept a lambda function taking one or two arguments and return an
# operator closure around it. These take the desired docstring as the second
# argument.

SINGLE_KEY_COMMANDS = collections.OrderedDict([
    ("\x7f", _drop),
    ("\x08", _drop),
    (curses.KEY_BACKSPACE, _drop),

    ("\r", _dup),
    ("\n", _dup),
    (curses.KEY_ENTER, _dup),

    (curses.KEY_UP, _rotup),
    (curses.KEY_DOWN, _rotdown),

    ("-", BINARY(lambda x, y: x - y, "subtract")),
    ("+", BINARY(lambda x, y: x + y, "add")),
    ("*", BINARY(lambda x, y: x * y, "multiply")),
    ("/", BINARY(lambda x, y: x / y, "divide")),
    ("%", BINARY(lambda x, y: x % y, "modulo")),
    ("^", BINARY(lambda x, y: x ** y, "power")),
    ("E", UNARY(autoc("exp"), "exponential")),
    ("l", UNARY(autoc("log"), "log base e")),
    ("L", UNARY(autoc("log10"), "log base 10")),
    ("r", UNARY(lambda x: x ** 0.5, "sq root")),
    ("i", UNARY(lambda x: 1/x, "reciprocal")),
    ("!", UNARY(lambda x: math.gamma(x + 1), "factorial")),
    ("P", BINARY(lambda x, y: math.gamma(x+1) / math.gamma(x-y+1), "permutations")),
    ("C", BINARY(lambda x, y: math.gamma(x+1) / (math.gamma(y+1)*math.gamma(x-y+1)), "combinations")),

    ("u", _undo),
    ("R", _redo),
    ("x", _xchg),
    ("[", _get),
    ("{", _pull),
    ("}", _push),

    ("t", _tag),
    ("T", _untag),

    ("S", _newstack),
    (">", _enterstack),
    ("<", _leavestack),

    ("?", _help),

    # RESERVED: these appear inside numbers
    ("e", None),
    ("j", None),
    ("h", None),
    ("d", None),
    ("o", None),
    ("b", None),
    ('"', None),
    ("_", None),
    (".", None),
])

################################################################################
# ADDING NAMED COMMANDS
#
# Insert a tuple into the OrderedDict below:
# (key, command)
#
# name:     the full command name, including beginning single-quote
# command:  a function that takes the current UndoStack, operates on it,
#               and returns nothing. Any exceptions thrown in here will be
#               caught and handled, so don't worry about those.
#
# The command must have a docstring. The first line of this will be used as its
# help text.
#
# Standard unary and binary operators can be made with UNARY() and BINARY(),
# which accept a lambda function taking one or two arguments and return an
# operator closure around it. These take the desired docstring as the second
# argument.
#
# You can insert section headers into the help text by adding a tuple:
# (id, header)
#
# id:       any unique integer
# header:   header text

COMMANDS = collections.OrderedDict([
    (10, "==TRIGONOMETRY, RAD=="),
    ("'sin", UNARY(autoc("sin"), "sin(rad)")),
    ("'cos", UNARY(autoc("cos"), "cos(rad)")),
    ("'tan", UNARY(autoc("tan"), "tan(rad)")),
    ("'asin", UNARY(autoc("asin"), "asin->rad")),
    ("'acos", UNARY(autoc("acos"), "acos->rad")),
    ("'atan", UNARY(autoc("atan"), "atan->rad")),
    (20, "==TRIGONOMETRY, DEG=="),
    ("'sind", UNARY(degree_trig(autoc("sin")), "sin(deg)")),
    ("'cosd", UNARY(degree_trig(autoc("cos")), "cos(deg)")),
    ("'tand", UNARY(degree_trig(autoc("tan")), "tan(deg)")),
    ("'asind", UNARY(degree_invtrig(autoc("asin")), "asin->deg")),
    ("'acosd", UNARY(degree_invtrig(autoc("acos")), "acos->deg")),
    ("'atand", UNARY(degree_invtrig(autoc("atan")), "atan->deg")),
    ("'deg", UNARY(lambda x: x * 180 / math.pi, "rad->deg")),
    ("'rad", UNARY(lambda x: x * math.pi / 180, "deg->rad")),
    (30, "==HYPERBOLICS=="),
    ("'sinh", UNARY(autoc("sinh"), "sinh")),
    ("'cosh", UNARY(autoc("cosh"), "cosh")),
    ("'tanh", UNARY(autoc("tanh"), "tanh")),
    ("'asinh", UNARY(autoc("asinh"), "asinh")),
    ("'acosh", UNARY(autoc("acosh"), "acosh")),
    ("'atanh", UNARY(autoc("atanh"), "atanh")),
    (40, "==LOGARITHMS=="),
    ("'exp", UNARY(autoc("exp"), "exponential")),
    ("'log", UNARY(autoc("log"), "log base e")),
    ("'log10", UNARY(autoc("log10"), "log base 10")),
    ("'log2", UNARY(lambda x: autoc("log")(x, 2), "log base 2")),
    (50, "==COMPLEX=="),
    ("'re", UNARY(lambda x: x.real, "real part")),
    ("'im", UNARY(lambda x: x.imag, "imaginary part")),
    ("'conj", UNARY(lambda x: complex(x.real, -x.imag), "complex conjugate")),
    ("'reim", _reim),
    ("'abs", UNARY(abs, "absolute value")),
    ("'angle", UNARY(cmath.phase, "angle/phase (rad)")),
    ("'polar", _polar),
    ("'fcart", _fcart),
    ("'fpolar", _fpolar),
    (60, "==NESTED STACKS=="),
    ("'sum", UNARY(sum, "sum of stack", autostack=False)),
    ("'mean", UNARY(lambda x: sum(x) / float(len(x)), "mean of stack", autostack=False)),
    ("'rms", UNARY(lambda x: (sum(i**2 for i in x) / float(len(x)))**0.5, "quadratic mean of stack", autostack=False)),
    ("'pvar", _pvar),
    ("'pdev", _pdev),
    ("'svar", _svar),
    ("'sdev", _sdev),
    ("'merge", _merge),
    ("'flatten", _flatten),
    ("'seq", _seq),
    (70, "==CONSTANTS=="),
    ("'pi", CONST(math.pi, "const PI")),
    ("'e", CONST(math.e, "const E")),
    ("'j", CONST(0+1j, "imaginary unit")),
    ("'c", CONST(2.99792458e8, "speed of light, m/s")),
    ("'h", CONST(6.6260755e-34, "Planck constant, J s")),
    ("'k", CONST(1.380658e-23, "Boltzmann constant, J/K")),
    ("'elec", CONST(1.60217733e-19, "Charge of electron, C")),
    ("'e0", CONST(8.854187817e-12, "Permittivity of vacuum, F/m")),
    ("'amu", CONST(1.6605402e-27, "Atomic mass unit, kg")),
    ("'Na", CONST(6.0221367e23, "Avogadro's number, mol^-1")),
    ("'atm", CONST(101325., "Standard atmosphere, Pa")),
    (80, "==LERPN=="),
    ("'sh", _sh),
    ("'help", _help),
    ("'exc", _exc),
    ("'repr", _repr),
    ("'str", _str),
    ("'eng", _eng),
    ("'sci", _sci),
    ("'fix", _fix),
    ("'nat", _nat),
    ("'hex", _hex),
    ("'oct", _oct),
    ("'bin", _bin),
])



from setuptools import setup

import sys

setup(  name='lerpn',
        version='2.0',
        description='Linux? Engineering RPN calculator',
        url='https://github.com/cpavlina/lerpn',
        download_url = 'https://github.com/cpavlina/lerpn/tarball/2.0',
        author='Chris Pavlina',
        author_email='pavlina.chris@gmail.com',
        packages=['LerpnApp'],
        entry_points = {
            'console_scripts': [
                'lerpn = LerpnApp:main',
                ],
        },
        keywords = ['calculator', 'rpn'],
        classifiers = [
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console :: Curses",
            "Topic :: Utilities",
            "License :: Public Domain" ]
        )



"""Global environment

These variables are accessible to the entire lerpn instance.
"""

from . import nums

# Display format.
# One of the subclasses of DisplayMode from nums
FORMAT = nums.EngMode

# sys.exc_info() from the last exception caught
# This is used by the 'exc command to display the backtrace
LAST_ERROR = None

# The current stack object.
# Note that due to nested stack support, this can change from cycle to cycle.
STACK = None

# Curses STDSCR
STDSCR = None



"""Code dealing with numbers.
"""

import re
import math

FLOAT_RE = re.compile(r"[-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?")
FLOAT_RE_UNDER = re.compile(r"[-_+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?")

NUM_PARSE_RE = re.compile(
    r"\s*(?P<re>" + FLOAT_RE_UNDER.pattern + r")" +     # real part
    r"\s*(?P<im>[jJ]\s*" + FLOAT_RE_UNDER.pattern + ")?" + # imaginary part
    r'\s*(?P<tag>".*)?' +                               # tag
    "$")

# number with radix
RNUM_PARSE_RE = re.compile(
    r"\s*(?P<radix>0[bdoh])" +                          # radix
    r"\s*(?P<re>[-+]?[0-9A-Fa-f]+)" +                   # real part
    r'\s*(?P<tag>".*)?' +                               # tag
    "$")

def num_parser(in_str):
    """Parse a number from user input.

    This accepts any combination of real part, imaginary part, and tag;
    or any combination of radix, integer, and tag:

    1.2e3   j5.4e-8     " random complex number
    0h      fffe        " my favorite number :)
    """

    match = RNUM_PARSE_RE.match(in_str)
    if match is not None:
        # integer with radix
        return num_parser_with_radix(match)

    match = NUM_PARSE_RE.match(in_str)

    if match is None:
        raise ValueError("Cannot parse value %r" % in_str)

    real_str = match.group("re")
    if real_str.startswith("_"):
        real_str = "-" + real_str[1:]
    real = float(real_str)

    if match.group("im") is not None:
        imag_str = match.group("im")[1:].strip()
        if imag_str.startswith("_"):
            imag_str = "-" + imag_str[1:]
        imag = float(imag_str)
    else:
        imag = None

    if imag is None:
        value = real
    else:
        value = complex(real, imag)

    if match.group("tag") is not None:
        value = Tagged(value, match.group("tag")[1:].strip())

    return value

def num_parser_with_radix(match):
    """Take over from num_parser for anything with a radix."""

    radname = match.group("radix").lower()
    radix_map = {"0h": 16, "0d": 10, "0o": 8, "0b": 2}
    radix = radix_map[radname]

    value = float(int(match.group("re"), radix))

    if match.group("tag") is not None:
        value = Tagged(value, match.group("tag")[1:].strip())

    return value

class UndoStack(list):
    """Stack supporting an 'undo' action.
    This is used as the main RPN stack. You can append() and pop() and [] just
    like with a normal list. You can also use undopush() to push a duplicate of
    the stack itself onto a "stack of stacks", and then undo() to restore that.

    undo() moves the last item to the redo stack. Use drop() to drop it without
    doing this.
    """

    def __init__(self, v=None):
        """Initialize an UndoStack from an optional source list.
        """
        self.__redo_held = None
        if v is None:
            list.__init__(self)
            self.__stack = [[]]
            self.__redo = []

        else:
            list.__init__(self, v)
            self.__stack = [[]]
            self.__redo = []

        self.parent = None

    def __repr__(self):
        return 'UndoStack(%r)' % (list(self))

    def undopush(self):
        """Save the current stack state to be restored later with undo()
        """
        self.__redo_held = self.__redo
        self.__redo = []
        self.__stack.append(self[:])

    def undo(self):
        """Restore the last saved undo state"""
        self.__redo.append(self[:])
        self.drop()

    def drop(self):
        """Restore the last saved undo state, but do not push the current
        state to the redo stack.
        """
        if len(self.__stack) > 1:
            self[:] = self.__stack.pop()
        else:
            self[:] = []

    def redo(self):
        """Undo the last undo"""
        if self.__redo:
            self[:] = self.__redo.pop()

    def replace_redo(self):
        """replace the redo stack directly after undopush has cleared it."""
        self.__redo = self.__redo_held

    def __str__(self):
        return "<< STACK(%d) >>" % len(self)

    def __add__(self, other):
        if isinstance(other, UndoStack):
            assert self.parent is other.parent
            new_stack = UndoStack(list.__add__(self, other))
            new_stack.parent = self.parent
            return new_stack
        else:
            return self.__add_real__(other) # pylint: disable=no-member

    def __radd__(self, other):
        # pylint: disable=no-member
        return self.__radd_real__(other)

    @classmethod
    def generate_operators(cls):
        """Autogenerates a set of operator overloads"""
        # Add operations to Tagged
        ops = [
            ("add_real", lambda x, y: x+y),
            ("sub", lambda x, y: x-y),
            ("mul", lambda x, y: x*y),
            ("div", lambda x, y: x/y),
            ("truediv", lambda x, y: x/y),
            ("mod", lambda x, y: x%y),
            ("pow", lambda x, y: x**y),
            ]
        for opname, opfunc in ops:
            def method(self, other, opfunc=opfunc):
                """automatically defined operator"""
                if not isinstance(other, float) and not isinstance(other, complex):
                    return NotImplemented
                new_stack = cls(opfunc(i, other) for i in self)
                new_stack.parent = self.parent
                return new_stack

            def rmethod(self, other, opfunc=opfunc):
                """automatically defined operator"""
                if not isinstance(other, float) and not isinstance(other, complex):
                    return NotImplemented
                new_stack = cls(opfunc(other, i) for i in self)
                new_stack.parent = self.parent
                return new_stack

            setattr(cls, "__%s__" % opname, method)
            setattr(cls, "__r%s__" % opname, rmethod)

UndoStack.generate_operators()

class Tagged(object):
    """Tagged number object.
    This behaves like a number, but also contains a string tag. The values
    are accessible at .num and .tag
    """
    def __init__(self, num, tag):
        assert isinstance(tag, str)
        self._num = num
        self._tag = tag

    @property
    def num(self):
        """the number that is tagged"""
        return self._num

    @property
    def tag(self):
        """the tag applied to the number"""
        return self._tag

    def __repr__(self):
        return "Tagged(%r, %r)" % (self._num, self._tag)

    @classmethod
    def generate_operators(cls):
        """Autogenerates a set of operator overloads"""
        # Add operations to Tagged
        ops = [
            ("add", lambda x, y: x+y),
            ("sub", lambda x, y: x-y),
            ("mul", lambda x, y: x*y),
            ("div", lambda x, y: x/y),
            ("truediv", lambda x, y: x/y),
            ("mod", lambda x, y: x%y),
            ("pow", lambda x, y: x**y),
            ]
        for opname, opfunc in ops:
            # For the uninitiated, the default-argument parameters create a new
            # scope for the variable, allowing passing each loop iteration's value
            # to the closure instead of closing around the single final value.
            def method(self, other, opfunc=opfunc):
                """automatically defined operator"""
                if isinstance(other, cls):
                    # If both are tagged, just remove the tag.
                    return opfunc(self.num, other.num)
                if not isinstance(other, float) and not isinstance(other, complex):
                    return NotImplemented
                return cls(opfunc(self.num, other), self.tag)

            def rmethod(self, other, opfunc=opfunc):
                """automatically defined operator"""
                if isinstance(other, cls):
                    # If both are tagged, just remove the tag.
                    return opfunc(other.num, self.num)
                if not isinstance(other, float) and not isinstance(other, complex):
                    return NotImplemented
                return cls(opfunc(other, self.num), self.tag)

            setattr(cls, "__%s__" % opname, method)
            setattr(cls, "__r%s__" % opname, rmethod)

Tagged.generate_operators()

LUT = {
    -24: "y",
    -21: "z",
    -18: "a",
    -15: "f",
    -12: "p",
    -9: "n",
    -6: u"\xb5",
    -3: "m",
    0: "",
    3: "k",
    6: "M",
    9: "G",
    12: "T",
    15: "P",
    18: "E",
    21: "Z",
    24: "Y"
}
def eng(num, sigfigs=7):
    """Return num in engineering notation"""

    if isinstance(num, complex):
        real = num.real
        imag = num.imag
        imag_sign = "+" if imag >= 0.0 else "-"
        imag_abs = abs(imag)

        return "(%s %sj %s)" %(eng(real), imag_sign, eng(imag_abs))

    if num == 0.0:
        return "0"

    elif num < 0.0:
        return '-'+eng(-num, sigfigs)

    magnitude = math.floor(math.log10(num)/3)*3
    coeff = "{0:.{1}g}".format(num * 10.0 ** (-magnitude), sigfigs)
    lut = LUT.get(magnitude, None)
    if lut is None:
        return coeff + "e" + str(magnitude)
    if lut != '':
        lut = '_' + lut
    return coeff + lut

class DisplayMode(object):
    """Static class with methods to implement a display mode"""

    @classmethod
    def format_real(cls, num):
        """Format a real number (float) according to this display mode.
        Must be implemented for each mode."""
        raise NotImplementedError

    @classmethod
    def format(cls, num):
        """Format the number in the subclass mode. Do not call this directly
        on DisplayMode, only on its subclasses."""
        if isinstance(num, Tagged):
            return cls.format(num.num)

        elif isinstance(num, float):
            return cls.format_real(num)

        elif isinstance(num, complex):
            real = num.real
            imag = num.imag
            imag_sign = "+" if imag >= 0.0 else "-"
            imag_abs = abs(imag)

            return "(%s %sj %s)" % (cls.format_real(real), imag_sign, cls.format_real(imag_abs))

        elif isinstance(num, str):
            return repr(num)

        else:
            return str(num)

class EngMode(DisplayMode):
    """Engineering mode. Numbers are displayed similarly to scientific notation,
    but with exponents always a multiple of 3, and using metric prefixes:

    13000.  becomes 13_k
    """

    sigfigs = 7

    @classmethod
    def format_real(cls, num):
        """Format a float in engineering mode"""
        return eng(num, cls.sigfigs)

class FixMode(DisplayMode):
    """Fixed mode. Numbers are displayed with a fixed number of places after
    the decimal point."""

    digits = 6

    @classmethod
    def format_real(cls, num):
        """Format a float in fixed mode"""
        return "%.*f" % (cls.digits, num)

class NaturalMode(DisplayMode):
    """Natural mode. Very large or small numbers are displayed in scientific notation;
    midrange numbers are displayed in standard notation."""

    @classmethod
    def format_real(cls, num):
        """Format a float in natural mode"""
        return "%g" % num

class SciMode(DisplayMode):
    """Scientific mode. All numbers are displayed as as [-]d.ddde+dd"""

    @classmethod
    def format_real(cls, num):
        """Format a float in scientific mode"""
        return "%e" % num

class HexMode(DisplayMode):
    """Display integers in hexadecimal, floats in natural mode."""
    @classmethod
    def format_real(cls, num):
        """Format a float, rounded to integer, in hexadecimal"""

        def hex_h(v):
            """same as hex(), but use 0h as the prefix instead of 0x"""
            return "0h" + hex(v)[2:]

        return format_int(num, hex_h)

class OctMode(DisplayMode):
    """Display integers in octal, floats in natural mode."""
    @classmethod
    def format_real(cls, num):
        """Format a float, rounded to integer, in octal"""

        def consistent_oct(n):
            """Format octal consistently, always 0o123 rather than
            0123 regardless of Python version."""
            octalized = oct(n)
            if octalized.startswith("0o"):
                return octalized
            else:
                return "0o" + octalized[1:]

        return format_int(num, consistent_oct)

class BinMode(DisplayMode):
    """Display integers in binary, floats in natural mode."""
    @classmethod
    def format_real(cls, num):
        """Format a float, rounded to integer, in binary"""
        return format_int(num, bin)

def format_int(value, fxn):
    """If 'value', as a float, is "close" to an integer, format it as
    fxn(int(value)), otherwise as "%g" % value."""

    fracpart = value - float(int(value))
    if abs(fracpart) < 0.001 or abs(fracpart) > 0.999:
        return fxn(int(value))
    else:
        return "%g" % value
